/*
 *  AREAS.H
 *
 *  Released to the public domain.
 *
 *  Prototypes for AREAS.C.
 */

#ifndef __AREAS_H__
#define __AREAS_H__

int mainArea(void);
int selectarea(char *topMsg, int def);
void BuildList(char ***lst);

#endif
