/*
 *  MISC.H
 *
 *  Released to the public domain.
 *
 *  Prototypes for MISC.C.
 */

#ifndef __MISC_H__
#define __MISC_H__

void change_curr_addr(void);
void change_nodelist(void);
void change_username(void);
void set_switch(void);
void show_help(void);
int handle_rot(int c);
void sel_chs(void);
int sel_group(void);

#endif
