/*
 *  TEMPLATE.H
 *
 *  Released to the public domain.
 *
 *  Prototypes for TEMPLATE.C.
 */

#ifndef __TEMPLATE_H__
#define __TEMPLATE_H__

int MakeTemplateMsg(msg * m, msg * oldmsg, int olda, int type);
void make_tearline(char *buf);

#endif
