/*
 *  KEYCODE.H
 *
 *  Written by either jim nutt or John Dennis.  Changes by Paul Edwards
 *  and Andrew Clarke.  Released to the public domain.
 *
 *  Prototype for KEYCODE.C.
 */

#ifndef __KEYCODE_H__
#define __KEYCODE_H__

void keycode(void);

#endif
