/*
 *  KEYS.H
 *
 *  Written on 10-Jul-94 by John Dennis and released to the public domain.
 *
 *  Definitions for various keyboard combinations.
 */

#ifndef __KEYS_H__
#define __KEYS_H__

#define Key_C_F1   0x5e00
#define Key_C_F2   0x5f00
#define Key_C_F3   0x6000
#define Key_C_F4   0x6100
#define Key_C_F5   0x6200
#define Key_C_F6   0x6300
#define Key_C_F7   0x6400
#define Key_C_F8   0x6500
#define Key_C_F9   0x6600
#define Key_C_F10  0x6700
#define Key_C_Lft  0x7300
#define Key_C_Rgt  0x7400
#define Key_C_End  0x7500
#define Key_C_PgDn 0x7600
#define Key_C_Home 0x7700
#define Key_C_PgUp 0x8400
#define Key_C_F11  0x8900
#define Key_C_F12  0x8a00
#define Key_C_Up   0x8d00
#define Key_C_Dwn  0x9100
#define Key_C_Ins  0x9200
#define Key_C_Del  0x9300
#define Key_C_Tab  0x9400
#define Key_C_2    0x0300
#define Key_C_A    0x0001
#define Key_C_B    0x0002
#define Key_C_C    0x0003
#define Key_C_D    0x0004
#define Key_C_E    0x0005
#define Key_C_F    0x0006
#define Key_C_G    0x0007
#define Key_C_H    0x0008
#define Key_C_I    0x0009
#define Key_C_Ent  0x000a
#define Key_C_J    0x000a
#define Key_C_K    0x000b
#define Key_C_L    0x000c
#define Key_C_M    0x000d
#define Key_C_N    0x000e
#define Key_C_O    0x000f
#define Key_C_P    0x0010
#define Key_C_Q    0x0011
#define Key_C_R    0x0012
#define Key_C_S    0x0013
#define Key_C_T    0x0014
#define Key_C_U    0x0015
#define Key_C_V    0x0016
#define Key_C_W    0x0017
#define Key_C_X    0x0018
#define Key_C_Y    0x0019
#define Key_C_Z    0x001a
#define Key_C_6    0x000e
#define Key_C_BS   0x007f
#define Key_C_Brk  0x0003
#define Key_S_Tab  0x0f00
#define Key_S_F1   0x5400
#define Key_S_F2   0x5500
#define Key_S_F3   0x5600
#define Key_S_F4   0x5700
#define Key_S_F5   0x5800
#define Key_S_F6   0x5900
#define Key_S_F7   0x5a00
#define Key_S_F8   0x5b00
#define Key_S_F9   0x5c00
#define Key_S_F10  0x5d00
#define Key_S_F11  0x8700
#define Key_S_F12  0x8800
#define Key_S_Q    0x0051
#define Key_S_W    0x0057
#define Key_S_E    0x0045
#define Key_S_R    0x0052
#define Key_S_T    0x0054
#define Key_S_Y    0x0059
#define Key_S_U    0x0055
#define Key_S_I    0x0049
#define Key_S_O    0x004f
#define Key_S_P    0x0050
#define Key_S_A    0x0041
#define Key_S_S    0x0053
#define Key_S_D    0x0044
#define Key_S_F    0x0046
#define Key_S_G    0x0047
#define Key_S_H    0x0048
#define Key_S_J    0x004a
#define Key_S_K    0x004b
#define Key_S_L    0x004c
#define Key_S_Z    0x005a
#define Key_S_X    0x0058
#define Key_S_C    0x0043
#define Key_S_V    0x0056
#define Key_S_B    0x0042
#define Key_S_N    0x004e
#define Key_S_M    0x004d
#define Key_A_Esc  0x0100
#define Key_A_BS   0x0e00
#define Key_A_Q    0x1000
#define Key_A_W    0x1100
#define Key_A_E    0x1200
#define Key_A_R    0x1300
#define Key_A_T    0x1400
#define Key_A_Y    0x1500
#define Key_A_U    0x1600
#define Key_A_I    0x1700
#define Key_A_O    0x1800
#define Key_A_P    0x1900
#define Key_A_Ent  0x1c00
#define Key_A_A    0x1e00
#define Key_A_S    0x1f00
#define Key_A_D    0x2000
#define Key_A_F    0x2100
#define Key_A_G    0x2200
#define Key_A_H    0x2300
#define Key_A_J    0x2400
#define Key_A_K    0x2500
#define Key_A_L    0x2600
#define Key_A_Z    0x2c00
#define Key_A_X    0x2d00
#define Key_A_C    0x2e00
#define Key_A_V    0x2f00
#define Key_A_B    0x3000
#define Key_A_N    0x3100
#define Key_A_M    0x3200
#define Key_A_F1   0x6800
#define Key_A_F2   0x6900
#define Key_A_F3   0x6a00
#define Key_A_F4   0x6b00
#define Key_A_F5   0x6c00
#define Key_A_F6   0x6d00
#define Key_A_F7   0x6e00
#define Key_A_F8   0x6f00
#define Key_A_F9   0x7000
#define Key_A_F10  0x7100
#define Key_A_1    0x7800
#define Key_A_2    0x7900
#define Key_A_3    0x7a00
#define Key_A_4    0x7b00
#define Key_A_5    0x7c00
#define Key_A_6    0x7d00
#define Key_A_7    0x7e00
#define Key_A_8    0x7f00
#define Key_A_9    0x8000
#define Key_A_0    0x8100
#define Key_A_F11  0x8b00
#define Key_A_F12  0x8c00
#define Key_F1     0x3b00
#define Key_F2     0x3c00
#define Key_F3     0x3d00
#define Key_F4     0x3e00
#define Key_F5     0x3f00
#define Key_F6     0x4000
#define Key_F7     0x4100
#define Key_F8     0x4200
#define Key_F9     0x4300
#define Key_F10    0x4400
#define Key_Home   0x4700
#define Key_Up     0x4800
#define Key_PgUp   0x4900
#define Key_Lft    0x4b00
#define Key_Cent   0x4c00
#define Key_Rgt    0x4d00
#define Key_End    0x4f00
#define Key_Dwn    0x5000
#define Key_PgDn   0x5100
#define Key_Ins    0x5200
#define Key_Del    0x5300
#define Key_F11    0x8500
#define Key_F12    0x8600
#define Key_BS     0x0008
#define Key_Tab    0x0009
#define Key_Ent    0x000d
#define Key_Esc    0x001b
#define Key_Dot    0x002e
#define Key_Spc    0x0020
#define Key_Q      0x0071
#define Key_W      0x0077
#define Key_E      0x0065
#define Key_R      0x0072
#define Key_T      0x0074
#define Key_Y      0x0079
#define Key_U      0x0075
#define Key_I      0x0069
#define Key_O      0x006f
#define Key_P      0x0070
#define Key_A      0x0061
#define Key_S      0x0073
#define Key_D      0x0064
#define Key_F      0x0066
#define Key_G      0x0067
#define Key_H      0x0068
#define Key_J      0x006a
#define Key_K      0x006b
#define Key_L      0x006c
#define Key_Z      0x007a
#define Key_X      0x0078
#define Key_C      0x0063
#define Key_V      0x0076
#define Key_B      0x0062
#define Key_N      0x006e
#define Key_M      0x006d

#endif
