/*
 *  ECHOTOSS.H
 *
 *  Written on 28-Jun-97 by Andrew Clarke and released to the public domain.
 *
 *  Code to append to Msged ECHOTOSS.LOG file.
 */

#ifndef __ECHOTOSS_H__
#define __ECHOTOSS_H__

void echotoss_add(AREA *a);

#endif
